"""Tests for the s2s.rgdr module."""
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from lilio import Calendar
from lilio import resample
from s2spy import RGDR
from s2spy.rgdr import rgdr
from s2spy.rgdr import utils


TEST_FILE_PATH = "./tests/test_rgdr/test_data"

"""Setting this matplotlib setting is required for running tests on windows. If this
option is not set, matplotlib tries to use an interactive backend, which will result in
a "_tkinter.TclError"."""
matplotlib.use("Agg")

# pylint: disable=protected-access


@pytest.fixture(scope="class")
def dummy_calendar():
    cal = Calendar(anchor="08-01")
    cal.add_intervals("target", length="1M", n=4)
    cal.add_intervals("precursor", length="1M", n=4)
    return cal


@pytest.fixture(scope="module")
def raw_target():
    return xr.open_dataset(f"{TEST_FILE_PATH}/tf5_nc5_dendo_80d77.nc").sel(cluster=3)


@pytest.fixture(scope="module")
def raw_field():
    return xr.open_dataset(
        f"{TEST_FILE_PATH}/sst_daily_1979-2018_5deg_Pacific_175_240E_25_50N.nc"
    )


@pytest.fixture(scope="class")
def example_precursor_field(raw_field, dummy_calendar):
    cal = dummy_calendar.map_to_data(raw_field)
    return resample(cal, raw_field).sst


@pytest.fixture(scope="class")
def example_target_timeseries(raw_target, raw_field, dummy_calendar):
    cal = dummy_calendar.map_to_data(raw_field)
    return resample(cal, raw_target).ts


@pytest.fixture(scope="class")
def example_corr(example_target_timeseries, example_precursor_field):
    return rgdr.correlation(
        example_precursor_field.sel(i_interval=-4),
        example_target_timeseries.sel(i_interval=1),
        corr_dim="anchor_year",
    )


@pytest.fixture(scope="class")
def expected_labels():
    return np.array(
        [
            [0, 0, 0, -1, -1, 0, 0, -2, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, -2, -2, -2, -2, -2, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, -2, -2, -2, -2, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )


class TestUtils:
    """Validates the various utility functions used in RGDR"""

    testdata_spherical_area = [
        (-1, 2, 2, 49500),
        (45, 5, 5, 218500),
        (45, 5, 2.5, 109250),
    ]  # lat, dlat, dlon, area

    @pytest.mark.parametrize("lat, dlat, dlon, expected", testdata_spherical_area)
    def test_spherical_area(self, lat, dlat, dlon, expected):
        np.testing.assert_allclose(
            rgdr.spherical_area(lat, dlat, dlon), expected, rtol=0.05
        )

    testdata_intervals = [  # target_intervals, lag, precursor_intervals
        ([2], 1, [1]),
        ([1], 1, [-1]),
        ([-1], 1, [-2]),
        ([1, 2, 3, 4], 4, [-4, -3, -2, -1]),
    ]

    @pytest.mark.parametrize("intervals, lag, expected", testdata_intervals)
    def test_intervals_subtract(self, intervals, lag, expected):
        assert utils.intervals_subtract(intervals, lag) == expected

    def test_intervals_subtract_neg(self):
        with pytest.raises(ValueError):
            utils.intervals_subtract([1], -1)


class TestCorrelation:
    """Validates that the correlation ufunc works, as well as the input checks."""

    @pytest.fixture
    def dummy_dataarray(self):
        time = pd.date_range("20161001", "20211001", freq="60d")

        return xr.DataArray(
            np.tile(np.arange(len(time)), (2, 2, 1)),
            dims=["lat", "lon", "time"],
            coords={
                "time": time,
                "lat": np.arange(0, 2),
                "lon": np.arange(0, 2),
            },
        )

    @pytest.fixture
    def dummy_timeseries(self, dummy_dataarray):
        return dummy_dataarray.isel(lat=0, lon=0).drop_vars(["lat", "lon"])

    def test_pearsonr(self):
        result = rgdr._pearsonr_nan([0, 0, 1], [0, 1, 1])  # type: ignore
        expected = (0.5, 0.66666667)
        np.testing.assert_array_almost_equal(result, expected)

    def test_pearsonr_nan(self):
        result = rgdr._pearsonr_nan([np.nan, 0, 1], [0, 1, 1])  # type: ignore
        expected = (np.nan, np.nan)
        np.testing.assert_array_almost_equal(result, expected)

    def test_correlation(self, dummy_dataarray, dummy_timeseries):
        c_val, p_val = rgdr.correlation(dummy_dataarray, dummy_timeseries)

        np.testing.assert_equal(c_val.values, 1)
        np.testing.assert_equal(p_val.values, 0)

    def test_correlation_dim_name(self, dummy_dataarray, dummy_timeseries):
        da = dummy_dataarray.rename({"time": "i_interval"})
        ts = dummy_timeseries.rename({"time": "i_interval"})
        c_val, p_val = rgdr.correlation(da, ts, corr_dim="i_interval")

        np.testing.assert_equal(c_val.values, 1)
        np.testing.assert_equal(p_val.values, 0)

    def test_correlation_wrong_target_dim_name(self, dummy_dataarray, dummy_timeseries):
        ts = dummy_timeseries.rename({"time": "dummy"})
        with pytest.raises(AssertionError):
            rgdr.correlation(dummy_dataarray, ts)

    def test_correlation_wrong_field_dim_name(self, dummy_dataarray, dummy_timeseries):
        da = dummy_dataarray.rename({"time": "dummy"})
        with pytest.raises(AssertionError):
            rgdr.correlation(da, dummy_timeseries)

    def test_correlation_wrong_target_dims(self, dummy_dataarray):
        ts = dummy_dataarray.rename({"lat": "latitude"})
        with pytest.raises(AssertionError):
            rgdr.correlation(dummy_dataarray, ts)

    def test_partial_correlation(self):
        with pytest.raises(NotImplementedError):
            rgdr.partial_correlation("field", "target", "z")

    def test_regression(self):
        with pytest.raises(NotImplementedError):
            rgdr.regression("field", "target")


class TestDBSCAN:
    """Separately validates that the DBSCAN-based clustering works as expected."""

    @pytest.fixture
    def dummy_dbscan_params(self):
        return {"alpha": 0.025, "eps": 600, "min_area": None}

    def test_dbscan(
        self,
        example_corr,
        example_precursor_field,
        dummy_dbscan_params,
        expected_labels,
    ):
        corr, p_val = example_corr

        clusters = rgdr.masked_spherical_dbscan(
            example_precursor_field, corr, p_val, dummy_dbscan_params
        )

        np.testing.assert_array_equal(clusters["cluster_labels"], expected_labels)

    def test_dbscan_min_area(
        self,
        example_corr,
        example_precursor_field,
        dummy_dbscan_params,
        expected_labels,
    ):
        corr, p_val = example_corr
        dbscan_params = dummy_dbscan_params
        dbscan_params["min_area"] = 1000**2
        clusters = rgdr.masked_spherical_dbscan(
            example_precursor_field, corr, p_val, dbscan_params
        )

        expected_labels[expected_labels == -1] = 0  # Small -1 cluster is missing

        np.testing.assert_array_equal(clusters["cluster_labels"], expected_labels)


rgdr_test_config = {
    "target_intervals": [1],
    "lag": 4,
    "eps_km": 600,
    "alpha": 0.025,
    "min_area_km2": 0,
}


class TestRGDR:
    """Test RGDR and its methods."""

    @pytest.fixture
    def dummy_rgdr(self):
        return RGDR(**rgdr_test_config)  # type: ignore

    def test_init(self):
        rgdr = RGDR(**rgdr_test_config)  # type: ignore
        assert isinstance(rgdr, RGDR)

    def test_target_intervals(self, dummy_rgdr):
        assert dummy_rgdr.target_intervals == [1]

    def test_precursor_intervals(self, dummy_rgdr):
        assert dummy_rgdr.precursor_intervals == [-4]

    def test_repr(self, dummy_rgdr):
        eval(repr(dummy_rgdr))  # pylint: disable=eval-used

    @pytest.mark.parametrize("prop", ("cluster_map", "pval_map", "corr_map"))
    def test_fitted_properties_err(self, dummy_rgdr, prop):
        with pytest.raises(ValueError):
            getattr(dummy_rgdr, prop)

    def test_transform_before_fit(self, dummy_rgdr, example_precursor_field):
        "Should fail as RGDR first has to be fit to (training) data."
        with pytest.raises(ValueError):
            dummy_rgdr.transform(example_precursor_field)

    def test_fit(self, dummy_rgdr, example_precursor_field, example_target_timeseries):
        dummy_rgdr.fit(example_precursor_field, example_target_timeseries)
        assert dummy_rgdr._area is not None

    def test_transform(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries
    ):
        dummy_rgdr.fit(example_precursor_field, example_target_timeseries)
        clustered_data = dummy_rgdr.transform(example_precursor_field)

        expected_labels = np.array([-2, -1], dtype="int16")
        np.testing.assert_array_equal(clustered_data["cluster_labels"], expected_labels)

    def test_fit_transform_fits(
        self, example_precursor_field, example_target_timeseries
    ):
        # Ensures that after fit_transform, the rgdr object is fit.
        rgdr = RGDR(**rgdr_test_config)  # type: ignore
        _ = rgdr.fit_transform(example_precursor_field, example_target_timeseries)
        assert rgdr._area is not None

    def test_fit_transform(self, example_precursor_field, example_target_timeseries):
        rgdr = RGDR(**rgdr_test_config)  # type: ignore
        clustered_data = rgdr.fit_transform(
            example_precursor_field, example_target_timeseries
        )
        expected_labels = np.array([-2, -1], dtype="int16")
        np.testing.assert_array_equal(clustered_data["cluster_labels"], expected_labels)

    @pytest.mark.parametrize("prop", ("cluster_map", "pval_map", "corr_map"))
    def test_fitted_properties(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries, prop
    ):
        dummy_rgdr.fit(example_precursor_field, example_target_timeseries)
        getattr(dummy_rgdr, prop)

    def test_corr_preview(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries
    ):
        dummy_rgdr.preview_correlation(
            example_precursor_field, example_target_timeseries
        )

    def test_corr_plot_ax(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries
    ):
        _, (ax1, ax2) = plt.subplots(ncols=2)
        dummy_rgdr.preview_correlation(
            example_precursor_field, example_target_timeseries, ax1=ax1, ax2=ax2
        )

    def test_cluster_plot(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries
    ):
        dummy_rgdr.preview_clusters(example_precursor_field, example_target_timeseries)

    def test_cluster_plot_ax(
        self, dummy_rgdr, example_precursor_field, example_target_timeseries
    ):
        _, ax = plt.subplots()
        dummy_rgdr.preview_clusters(
            example_precursor_field, example_target_timeseries, ax=ax
        )

    def test_multiple_targets(self, example_precursor_field, example_target_timeseries):
        rgdr = RGDR(
            target_intervals=[1, 2], lag=4, eps_km=600, alpha=0.025, min_area_km2=0
        )

        clustered_data = rgdr.fit_transform(
            example_precursor_field, example_target_timeseries
        )

        expected_labels = np.array([-2, -1], dtype="int16")
        np.testing.assert_array_equal(clustered_data["cluster_labels"], expected_labels)



"""Response Guided Dimensionality Reduction."""
import warnings
from os import linesep
from typing import Optional
from typing import TypeVar
from typing import Union
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.collections import QuadMesh
from scipy.stats import pearsonr as _pearsonr
from sklearn.cluster import DBSCAN
from . import utils


RADIUS_EARTH_KM = 6371
SURFACE_AREA_EARTH_KM2 = 5.10072e8
XrType = TypeVar("XrType", xr.DataArray, xr.Dataset)


def spherical_area(latitude: float, dlat: float, dlon: Optional[float] = None) -> float:
    """Approximate the area of a square grid cell on a spherical (!) earth.

    Returns the area in square kilometers of earth surface.

    Args:
        latitude (float): Latitude at the center of the grid cell (deg)
        dlat (float): Latitude grid resolution (deg)
        dlon (float): Longitude grid resolution (deg), optional in case of a square grid.

    Returns:
        float: Area of the grid cell (km^2)
    """
    if dlon is None:
        dlon = dlat
    dlon = np.radians(dlon)
    dlat = np.radians(dlat)

    lat = np.radians(latitude)
    h = np.sin(lat + dlat / 2) - np.sin(lat - dlat / 2)
    spherical_area = h * dlon / (np.pi * 4)

    return spherical_area * SURFACE_AREA_EARTH_KM2


def cluster_area(ds: XrType, cluster_label: float) -> float:
    """Determine the total area of a cluster.

    Requires the input dataset to have the variables `area` and `cluster_labels`.

    Args:
        ds (xr.Dataset or xr.DataArray): Dataset/DataArray containing the variables
            `area` and `cluster_labels`.
        cluster_label (float): The label (as float) for which the area should be
            calculated.

    Returns:
        float: Area of the cluster `cluster_label`.
    """
    return (
        ds["area"]
        .where(ds["cluster_labels"] == cluster_label)
        .sum(skipna=True)
        .values.item()
    )


def remove_small_area_clusters(ds: XrType, min_area_km2: float) -> XrType:
    """Remove the clusters where the area is under the input threshold.

    Args:
        ds (xr.DataArray, xr.Dataset): Dataset containing `cluster_labels` and `area`.
        min_area_km2 (float): The minimum allowed area of each cluster

    Returns:
        xr.DataArray, xr.Dataset: The input dataset with the labels of the clusters set
            to 0 when the area of the cluster is under the `min_area_km2` threshold.
    """
    clusters = np.unique(ds["cluster_labels"])
    areas = [cluster_area(ds, c) for c in clusters]
    valid_clusters = np.array([c for c, a in zip(clusters, areas) if a > min_area_km2])

    ds["cluster_labels"] = ds["cluster_labels"].where(
        np.isin(ds["cluster_labels"], valid_clusters), 0
    )

    return ds


def add_gridcell_area(data: xr.DataArray):
    """Add the area of each gridcell (latitude) in km2.

    Note: Assumes an even grid (in degrees)

    Args:
        data: Data containing lat, lon coordinates in degrees.

    Returns:
        Input data with an added coordinate "area".
    """
    dlat = np.abs(data.latitude.values[1] - data.latitude.values[0])
    dlon = np.abs(data.longitude.values[1] - data.longitude.values[0])
    data["area"] = spherical_area(data.latitude, dlat, dlon)
    return data


def assert_clusters_present(data: xr.DataArray) -> None:
    """Assert that any (non-'0') clusters are present in the data."""
    if np.unique(data.cluster_labels).size == 1:
        warnings.warn(
            "No significant clusters found in the input DataArray", stacklevel=2
        )


def _get_dbscan_clusters(
    data: xr.Dataset, coords: np.ndarray, dbscan_params: dict
) -> np.ndarray:
    """Generate the DBSCAN cluster labels based on the correlation and p-value.

    Args:
        data: DataArray of the precursor field, of only a single
             i_interval. Requires the 'latitude' and 'longitude' dimensions to be stacked
             into a "coords" dimension.
        coords: 2-D array containing the coordinates of each (lat, lon) grid
            point, in radians.
        dbscan_params: Dictionary containing the elements 'alpha', 'eps',
            'min_area_km2'. See the documentation of RGDR for more information.

    Returns:
        np.ndarray: 1-D array of the same length as `coords`, containing cluster labels
            for every coordinate.
    """
    labels = np.zeros(len(coords), dtype=int)

    for sign, sign_mask in zip([1, -1], [data["corr"] >= 0, data["corr"] < 0]):
        mask = np.logical_and(data["p_val"] < dbscan_params["alpha"], sign_mask)

        if np.sum(mask) > 0:  # Check if the mask contains any points to cluster
            db = DBSCAN(
                eps=dbscan_params["eps"] / RADIUS_EARTH_KM,
                min_samples=1,
                algorithm="auto",
                metric="haversine",
            ).fit(coords[mask])

            labels[mask] = sign * (db.labels_ + 1)

    return labels


def _find_clusters(
    precursor: xr.DataArray,
    corr: xr.DataArray,
    p_val: xr.DataArray,
    dbscan_params: dict,
) -> xr.DataArray:
    """Compute clusters and adds their labels to the precursor dataset.

    For clustering the DBSCAN algorithm is used, with a Haversine distance metric.

    Args:
        precursor (xr.DataArray): DataArray of the precursor field, containing
            'latitude' and 'longitude' dimensions in degrees.
        corr (xr.DataArray): DataArray with the correlation values, generated by
            correlation_map()
        p_val (xr.DataArray): DataArray with the p-values, generated by
            correlation_map()
        dbscan_params (dict): Dictionary containing the elements 'alpha', 'eps',
            'min_area_km2'. See the documentation of RGDR for more information.

    Returns:
        xr.DataArray: The input precursor data, with as extra coordinate labelled
            clusters.
    """
    data = precursor.to_dataset()
    data["corr"], data["p_val"] = corr, p_val  # Will require less tracking of indices

    data = data.stack(coord=["latitude", "longitude"])
    coords = np.asarray(data["coord"].values.tolist())
    coords = np.radians(coords)

    labels = _get_dbscan_clusters(data, coords, dbscan_params)

    precursor = precursor.stack(coord=["latitude", "longitude"])
    precursor["cluster_labels"] = ("coord", labels)
    precursor["cluster_labels"] = precursor["cluster_labels"].astype("int16")
    precursor = precursor.unstack("coord")

    return precursor


def masked_spherical_dbscan(
    precursor: xr.DataArray,
    corr: xr.DataArray,
    p_val: xr.DataArray,
    dbscan_params: dict,
) -> xr.DataArray:
    """Determine the clusters based on sklearn's DBSCAN implementation.

    Alpha determines the mask based on the minimum p_value. Grouping can be
    adjusted using the `eps_km` parameter. Cluster labels are negative for
    areas with a negative correlation coefficient and positive for areas with
    a positive correlation coefficient. Areas without any significant correlation
    are put in the cluster labelled '0'.

    Args:
        precursor (xr.DataArray): DataArray of the precursor field, containing
            'latitude' and 'longitude' dimensions in degrees.
        corr (xr.DataArray): DataArray with the correlation values, generated by
            correlation_map()
        p_val (xr.DataArray): DataArray with the p-values, generated by
            correlation_map()
        dbscan_params (dict): Dictionary containing the elements 'alpha', 'eps',
            'min_area_km2'. See the documentation of RGDR for more information.

    Returns:
        xr.DataArray: Precursor data grouped by the DBSCAN clusters.
    """
    precursor = add_gridcell_area(precursor)
    precursor = _find_clusters(precursor, corr, p_val, dbscan_params)

    if dbscan_params["min_area"] is not None:
        precursor = remove_small_area_clusters(precursor, dbscan_params["min_area"])

    assert_clusters_present(precursor)

    return precursor


def _pearsonr_nan(x: np.ndarray, y: np.ndarray) -> tuple[float, float]:
    """NaN friendly implementation of scipy.stats.pearsonr.

    Calculates the correlation coefficient between two arrays, as well as the p-value
    of this correlation. However, instead of raising an error when encountering NaN
    values, this function will return both the correlation coefficient and the p-value as NaN.

    Args:
        x: 1-D array
        y: 1-D array
    Returns:
        r_coefficient
        p_value

    """
    if np.any(np.isnan(x)) or np.any(np.isnan(y)):
        return np.nan, np.nan
    return _pearsonr(x, y)


def correlation(
    field: xr.DataArray, target: xr.DataArray, corr_dim: str = "time"
) -> tuple[xr.DataArray, xr.DataArray]:
    """Calculate correlation maps.

    Args:
        field: Spatial data with a dimension named `corr_dim`, over which each
            location should have the Pearson correlation coefficient calculated with the
            target data.
        target: Data which has to be correlated with the spatial data. Requires a
            dimension named `corr_dim`.
        corr_dim: Dimension over which the correlation coefficient should be calculated.

    Returns:
        r_coefficient: DataArray filled with the correlation coefficient for each
            non-`corr_dim` coordinate.
        p_value: DataArray filled with the two-tailed p-values for each computed
            correlation coefficient.
    """
    assert (
        corr_dim in target.dims
    ), f"input target does not have contain the '{corr_dim}' dimension"
    assert (
        corr_dim in field.dims
    ), f"input field does not have contain the '{corr_dim}' dimension"
    assert np.all(
        [dim in field.dims for dim in target.dims]
    ), "Field and target dims do not match"

    return xr.apply_ufunc(
        _pearsonr_nan,
        field,
        target,
        input_core_dims=[[corr_dim], [corr_dim]],
        vectorize=True,
        output_core_dims=[[], []],
    )


def partial_correlation(field, target, z):
    """Calculate partial correlation maps."""
    raise NotImplementedError


def regression(field, target):
    """Regression analysis on entire maps.

    Methods include Linear, Ridge, Lasso.
    """
    raise NotImplementedError


def stack_input_data(precursor, target, precursor_intervals, target_intervals):
    """Stack input data."""
    target = target.sel(i_interval=target_intervals).stack(
        anch_int=["anchor_year", "i_interval"]
    )
    precursor = precursor.sel(i_interval=precursor_intervals).stack(
        anch_int=["anchor_year", "i_interval"]
    )

    precursor = precursor.drop_vars({"anchor_year", "anch_int", "i_interval"})
    target = target.drop_vars({"anchor_year", "anch_int", "i_interval"})

    precursor["anch_int"] = range(precursor["anch_int"].size)
    target["anch_int"] = range(target["anch_int"].size)

    return precursor, target


class RGDR:
    """Response Guided Dimensionality Reduction."""

    def __init__(  # noqa: PLR0913 (too-many-arguments)
        self,
        target_intervals: Union[int, list[int]],
        lag: int,
        eps_km: float,
        alpha: float,
        min_area_km2: Optional[float] = None,
    ) -> None:
        """Response Guided Dimensionality Reduction (RGDR).

        Dimensionality reduction based on the correlation between precursor field and
        target timeseries.

        Args:
            target_intervals: The target interval indices which should be correlated
                with the precursors. Input in the form of "[1, 2, 3]" or "1"
                The precursor intervals will be determined based on the `lag` kwarg.
            lag: The lag between the precursor and target intervals to compute the
                correlation, akin to lag in cross-correlation. E.g. if the target
                intervals are [1, 2], and lag is 2, the precursor intervals will be
                [-2, -1]
            alpha (float): p-value below which the correlation is considered significant
                enough for a location to be included in a cluster.
            eps_km (float): The maximum distance (in km) between two grid cells for them
                to be considered to be in the same cluster. This is not a maximum bound
                on the distances between grid cells within a cluster.

                The minimum appropriate value is equal to the maximum width/height of a
                grid cell (i.e. the width of the grid cell closest to the equator).
                The upper bound depends on when cells or clusters would still be
                considered to be part of the same climate signal (i.e., precursor region).

                Higher values can lead to fewer clusters, but also clusters in which the
                cells of the same cluster are separated by large geographical distances.
            min_area_km2 (float): The minimum area of a cluster (in square km). Clusters
                smaller than this minimum area will be discarded.

        Attributes:
            corr_map: correlation coefficient map of given precursor field and
                target series.
            pval_map: p-values map of correlation
            cluster_map: cluster labels for precursor field masked by p-values
        """
        self._lag = lag
        self._dbscan_params = {"eps": eps_km, "alpha": alpha, "min_area": min_area_km2}

        self._target_intervals = (
            [target_intervals]
            if isinstance(target_intervals, int)
            else target_intervals
        )
        self._precursor_intervals = utils.intervals_subtract(
            self._target_intervals, lag
        )

        self._corr_map: Union[None, xr.DataArray] = None
        self._pval_map: Union[None, xr.DataArray] = None
        self._cluster_map: Union[None, xr.DataArray] = None

        self._area = None

    @property
    def target_intervals(self) -> list[int]:
        """Return target intervals."""
        return self._target_intervals

    @property
    def precursor_intervals(self) -> list[int]:
        """Return precursor intervals."""
        return self._precursor_intervals

    @property
    def cluster_map(self) -> xr.DataArray:
        """Return cluster map."""
        if self._cluster_map is None:
            raise ValueError(
                "No cluster map exists yet, .fit() has to be called first."
            )
        return self._cluster_map

    @property
    def pval_map(self) -> xr.DataArray:
        """Return p-value map."""
        if self._pval_map is None:
            raise ValueError(
                "No p-value map exists yet, .fit() has to be called first."
            )
        return self._pval_map

    @property
    def corr_map(self) -> xr.DataArray:
        """Return correlation map."""
        if self._corr_map is None:
            raise ValueError(
                "No correlation map exists yet, .fit() has to be called first."
            )
        return self._corr_map

    def get_correlation(
        self,
        precursor: xr.DataArray,
        target: xr.DataArray,
    ) -> tuple[xr.DataArray, xr.DataArray]:
        """Calculate the correlation and p-value between input precursor and target.

        Args:
            precursor: Precursor field data with the dimensions
                'latitude', 'longitude', and 'anchor_year'
            target: Timeseries data with only the dimension 'anchor_year'

        Returns:
            (correlation, p_value): DataArrays containing the correlation and p-value.
        """
        if not isinstance(precursor, xr.DataArray):
            raise ValueError("Please provide an xr.DataArray, not a dataset")

        p, t = stack_input_data(
            precursor, target, self._precursor_intervals, self._target_intervals
        )

        return correlation(p, t, corr_dim="anch_int")

    def get_clusters(
        self,
        precursor: xr.DataArray,
        target: xr.DataArray,
    ) -> xr.DataArray:
        """Generate clusters for the precursor data.

        Args:
            precursor: Precursor field data with the dimensions
                'latitude', 'longitude', 'anchor_year', and 'i_interval'
            target: Target timeseries data with only the dimensions 'anchor_year' and
                'i_interval'

        Returns:
            DataArray containing the clusters as masks.
        """
        corr, p_val = self.get_correlation(precursor, target)
        return masked_spherical_dbscan(precursor, corr, p_val, self._dbscan_params)

    def preview_correlation(  # noqa: PLR0913 (too-many-arguments)
        self,
        precursor: xr.DataArray,
        target: xr.DataArray,
        add_alpha_hatch: bool = True,
        ax1: Optional[plt.Axes] = None,
        ax2: Optional[plt.Axes] = None,
    ) -> list[QuadMesh]:
        """Preview correlation and p-value results with given inputs.

        Generate a figure showing the correlation and p-value results with the
        initiated RGDR class and input precursor field.

        Args:
            precursor: Precursor field data with the dimensions
                'latitude', 'longitude', 'anchor_year', and 'i_interval'
            target: Target timeseries data with only the dimensions 'anchor_year' and
                'i_interval'
            add_alpha_hatch: Adds a red hatching when the p-value is lower than the
                RGDR's 'alpha' value.
            ax1: a matplotlib axis handle to plot the correlation values into.
                If None, an axis handle will be created instead.
            ax2: a matplotlib axis handle to plot the p-values into. If None, an axis
                handle will be created instead.

        Returns:
            List of matplotlib QuadMesh artists.
        """
        corr, p_val = self.get_correlation(precursor, target)

        if (ax1 is None) and (ax2 is None):
            _, (ax1, ax2) = plt.subplots(ncols=2)
        elif (ax1 is None) or (ax2 is None):
            raise ValueError(
                "Either pass axis handles for both ax1 and ax2, or pass neither."
            )

        plot1 = corr.plot.pcolormesh(ax=ax1, cmap="coolwarm")  # type: ignore
        plot2 = p_val.plot.pcolormesh(ax=ax2, cmap="viridis_r", vmin=0, vmax=1)  # type: ignore

        if add_alpha_hatch:
            coords = plot2.get_coordinates()
            plt.rcParams["hatch.color"] = "r"
            plt.pcolor(
                coords[:, :, 0],
                coords[:, :, 1],
                p_val.where(p_val < self._dbscan_params["alpha"]).values,
                hatch="x",
                alpha=0.0,
            )

        ax1.set_title("correlation")
        ax2.set_title("p-value")

        return [plot1, plot2]

    def preview_clusters(
        self,
        precursor: xr.DataArray,
        target: xr.DataArray,
        ax: Optional[plt.Axes] = None,
        **kwargs,
    ) -> QuadMesh:
        """Preview clusters.

        Generates a figure showing the clusters resulting from the initiated RGDR
        class and input precursor field.

        Args:
            precursor: Precursor field data with the dimensions
                'latitude', 'longitude', 'anchor_year', and 'i_interval'
            target: Target timeseries data with only the dimensions 'anchor_year' and
                'i_interval'
            ax (plt.Axes, optional): a matplotlib axis handle to plot the clusters
                into. If None, an axis handle will be created instead.
            **kwargs: Keyword arguments that should be passed to QuadMesh.

        Returns:
            Matplotlib QuadMesh artist.
        """
        if ax is None:
            _, ax = plt.subplots()

        clusters = self.get_clusters(precursor, target)

        return clusters["cluster_labels"].plot(cmap="viridis", ax=ax, **kwargs)  # type: ignore

    def fit(self, precursor: xr.DataArray, target: xr.DataArray):
        """Fit RGDR clusters to precursor data.

        Performs DBSCAN clustering on a prepared DataArray, and then groups the data by
        their determined clusters, using an weighted mean. The weight is based on the
        area of each grid cell.

        Density-Based Spatial Clustering of Applications with Noise (DBSCAN) clusters
        gridcells together which are of the same sign and in proximity to
        each other using DBSCAN.

        Clusters labelled with a positive value represent a positive correlation with
        the target timeseries, the clusters labelled with a negative value represent a
        negative correlation. All locations not in a cluster are grouped together under
        the label '0'.

        Args:
            precursor: Precursor field data with the dimensions
                'latitude', 'longitude', 'anchor_year', and 'i_interval'
            target: Target timeseries data with only the dimensions 'anchor_year' and
                'i_interval', which will be correlated with the precursor field.

        Returns:
            xr.DataArray: The precursor data, with the latitute and longitude dimensions
                reduced to clusters.
        """
        corr, p_val = self.get_correlation(precursor, target)

        masked_data = masked_spherical_dbscan(
            precursor, corr, p_val, self._dbscan_params
        )
        self._corr_map = corr
        self._pval_map = p_val
        self._cluster_map = masked_data.cluster_labels
        self._area = masked_data.area

    def transform(self, data: xr.DataArray) -> xr.DataArray:
        """Apply RGDR on the input data, based on the previous fit.

        Transform will use the clusters previously generated when RGDR was fit, and use
        these clusters to reduce the latitude and longitude dimensions of the input
        data.
        """
        if self.cluster_map is None:
            raise ValueError(
                "Transform requires the model to be fit on other data first"
            )
        data = data.sel(i_interval=self._precursor_intervals)
        data["cluster_labels"] = self.cluster_map
        data["area"] = self._area

        reduced_data = utils.weighted_groupby(
            data, groupby="cluster_labels", weight="area"
        )

        # Add the geographical centers for later alignment between, e.g., splits
        reduced_data = utils.geographical_cluster_center(data, reduced_data)
        # Include explanations about geographical centers as attributes
        reduced_data.attrs[
            "data"
        ] = "Clustered data with Response Guided Dimensionality Reduction."
        reduced_data.attrs[
            "coordinates"
        ] = "Latitudes and longitudes are geographical centers associated with clusters."

        # Remove the '0' cluster
        reduced_data = reduced_data.where(reduced_data["cluster_labels"] != 0).dropna(
            dim="cluster_labels"
        )

        return reduced_data.transpose(..., "cluster_labels")

    def fit_transform(self, precursor: xr.DataArray, timeseries: xr.DataArray):
        """Fit RGDR clusters to precursor data, and applies RGDR on the input data.

        Args:
            precursor: Precursor field data with the dimensions 'latitude', 'longitude',
                and 'anchor_year'
            timeseries: Timeseries data with only the dimension 'anchor_year', which
                will be correlated with the precursor field.

        Returns:
            xr.DataArray: The precursor data, with the latitute and longitude dimensions
                reduced to clusters.
        """
        self.fit(precursor, timeseries)
        return self.transform(precursor)

    def __repr__(self) -> str:
        """Represent the RGDR transformer with strings."""
        props = [
            ("target_intervals", repr(self.target_intervals)),
            ("lag", repr(self._lag)),
            ("eps_km", repr(self._dbscan_params["eps"])),
            ("alpha", repr(self._dbscan_params["alpha"])),
            ("min_area_km2", repr(self._dbscan_params["min_area"])),
        ]

        propstr = f"{linesep}\t" + f",{linesep}\t".join(
            [f"{k}={v}" for k, v in props]
        )  # sourcery skip: use-fstring-for-concatenation
        return f"{self.__class__.__name__}({propstr}{linesep})".replace("\t", "    ")



"""s2spy: integrating expert knowledge and ai to boost S2S forecasting.

This package is a high-level python package integrating expert knowledge
and artificial intelligence to boost (sub) seasonal forecasting.
"""
import logging
from .rgdr.rgdr import RGDR


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Yang Liu"
__email__ = "y.liu@esciencecenter.nl"
__version__ = "0.4.0"

__all__ = ["RGDR"]



"""Label alignment tools for RGDR clusters."""
import itertools
import string
from copy import copy
from copy import deepcopy
from typing import TYPE_CHECKING
import numpy as np
import pandas as pd
import xarray as xr


if TYPE_CHECKING:
    from s2spy.rgdr.rgdr import RGDR


def _get_split_cluster_dict(cluster_labels: xr.DataArray) -> dict:
    """Generate a dictionary of all cluster labels in each split.

    Args:
        cluster_labels: DataArray containing all the cluster maps, with the dimension
            "split" for the different clusters over splits.

    Returns:
        Dictionary in the form {0: [cluster_a, cluster_b], 1: [cluster_a], ...}
    """
    return {
        i_split: list(np.unique(split_data.values)[np.unique(split_data.values) != 0])
        for i_split, split_data in enumerate(cluster_labels)
    }


def _flatten_cluster_dict(cluster_dict: dict) -> list[tuple[int, int]]:
    """Flattens a cluster dictionary to a list with (split, cluster) as values.

    For example, if the input is {0: [-1, -2, 1], 1: [-1, 1]}, this function will return
    the following list: [(0, -1), (0, -2), (0, 1), (1, -1), (1, 1)]

    Args:
        cluster_dict: The cluster dictionary which should be flattened

    Returns:
        A list of the clusters and their splits
    """
    flat_clusters: list[tuple[int, int]] = []
    for split in cluster_dict:
        flat_clusters.extend((split, cluster) for cluster in cluster_dict[split])
    return flat_clusters


def _init_overlap_df(cluster_labels: xr.DataArray):
    """Build an empty dataframe with multi-indexes for clusters and labels.

    The structure will be something like the following table:

    split       |        0         1
    label       |       -1        -1
    ------------|-------------------
    split label |
    0     -1    |      NaN  0.583333
    1     -1    | 0.333333       NaN

    The same multi-index is used for both rows and columns, such that the dataframe can
    be populated with the overlap between labels from different splits.

    Args:
        cluster_labels: DataArray containing all the cluster maps, with the dimension
            "split" for the different clusters over splits.

    Returns:
        A pandas dataframe containing a table
    """
    split_label_dict = _get_split_cluster_dict(cluster_labels)
    flat_clusters = _flatten_cluster_dict(split_label_dict)
    multi_index = pd.MultiIndex.from_tuples(flat_clusters, names=("split", "label"))

    return pd.DataFrame(np.nan, index=multi_index, columns=multi_index)


def _calculate_overlap(
    cluster_labels: xr.DataArray,
    split_a: int,
    cluster_a: int,
    split_b: int,
    cluster_b: int,
) -> float:
    """Calculate the overlapping fraction between two clusters, over different splits.

    The overlap is defines as:
        overlap = n_overlapping_cells / total_cells_cluster_a

    Args:
        cluster_labels: DataArray containing all the cluster maps, with the dimension
            "split" for the different clusters over splits.
        split_a: The index of the split of the first cluster
        cluster_a: The value of the first cluster in the clusters_da DataArray.
        split_b: The index of the split of the second cluster
        cluster_b: The value of the second cluster in the clusters_da DataArray.

    Returns:
        Overlap of the first cluster with the second cluster, as a fraction (0.0 - 1.0)
    """
    mask_a = xr.where(cluster_labels.sel(split=split_a) == cluster_a, 1, 0).values
    mask_b = xr.where(cluster_labels.sel(split=split_b) == cluster_b, 1, 0).values

    return np.sum(np.logical_and(mask_a, mask_b)) / np.sum(mask_a)


def calculate_overlap_table(cluster_labels: xr.DataArray) -> pd.DataFrame:
    """Fill the overlap table with the overlap between clusters over different splits.

    Args:
        cluster_labels: DataArray containing all the cluster maps, with the dimension
            "split" for the different clusters over splits.

    Returns:
        The overlap table with all valid combinations filled in. Non valid combinations
            of clusters (the cluster itself, or within the same split) will have NaN
            values.
    """
    overlap_df = _init_overlap_df(cluster_labels)

    all_clusters = _get_split_cluster_dict(cluster_labels)

    pos_clusters = {
        split: [el for el in cluster_list if np.sign(el) == 1]
        for (split, cluster_list) in all_clusters.items()
    }
    neg_clusters = {
        split: [el for el in cluster_list if np.sign(el) == -1]
        for (split, cluster_list) in all_clusters.items()
    }

    for clusters in (pos_clusters, neg_clusters):
        flat_clusters = _flatten_cluster_dict(clusters)

        for split, cluster in flat_clusters:
            other_clusters = _flatten_cluster_dict(
                {k: v for (k, v) in clusters.items() if k != split}
            )
            for other_split, other_cluster in other_clusters:
                overlap = _calculate_overlap(
                    cluster_labels, split, cluster, other_split, other_cluster
                )
                overlap_df.at[(split, cluster), (other_split, other_cluster)] = overlap

    return overlap_df


def get_overlapping_clusters(
    cluster_labels: xr.DataArray, min_overlap: float = 0.1
) -> set:
    """Create sets of overlapping clusters.

    Clusters will be considered to have sufficient overlap if they overlap at least by
    the minimum threshold. Note that this is a one way criterion.

    For example, if the overlap table is like the following:
    split       |    0       1
    label       |   -1      -1
    ------------|-------------
    split label |
    0     -1    |  NaN    0.05
    1     -1    | 0.20     NaN

    Then cluster (split: 0, label: -1) will overlap with cluster (1, -1) by 0.05. This
    is insufficient to be considered the same cluster. However, cluster (1, -1) does
    overlap by 0.20 with cluster (0, -1), so they *will* be considered the same cluster.
    This situation can arise when one cluster is much bigger than another one.

    In this example, the overlapping set will be {frozenset("0_-1", "1_-1")}.
    Note that if we would use a threshold of 0.05, the output would not change, as the
    two nexted sets {"0_-1", "1_-1"} and {"1_-1", "0_-1"} are the same.

    Args:
        cluster_labels: DataArray containing all the cluster maps, with the dimension
            "split" for the different clusters over splits.
        min_overlap: Minimum overlap (0.0 - 1.0) when clusters are considered to be
            sufficiently overlapping to belong to the same signal. Defaults to 0.1.

    Returns:
        A set of (frozen) sets, each set corresponding to a possible combination of
            clusters that overlap.
    """
    overlap_df = calculate_overlap_table(cluster_labels)

    overlap_df.columns = ["_".join([str(el) for el in col]) for col in overlap_df.columns.values]  # type: ignore
    overlap_df.index = ["_".join([str(el) for el in idx]) for idx in overlap_df.index.values]  # type: ignore

    clusters = set()
    for row, _ in enumerate(overlap_df.index):
        overlapping = (
            overlap_df.iloc[row].where(overlap_df.iloc[row] > min_overlap).dropna()
        )
        cluster = set(overlapping.index)
        cluster.add(overlap_df.iloc[row].name)
        clusters.add(frozenset(cluster))

    return clusters


def remove_subsets(clusters: set) -> set:
    """Remove subsets from the clusters.

    For example: {{"A"}, {"A", "B"}} will become {{"A", "B"}}, as "A" is a subset of the
    bigger cluster.
    """
    no_subsets = set(clusters)
    for cluster in clusters:
        no_subsets.remove(cluster)
        if not any(cluster.issubset(cl) for cl in no_subsets):
            no_subsets.add(cluster)
    return no_subsets


def remove_overlapping_clusters(clusters: set) -> set:
    """Remove clusters shared between two different groups of clusters.

    Largest cluster gets priority.

    For example: {{"A", "D"}, {"A", "B", "C"}} will become {{"D"}, {"A", "B", "C"}}
    """
    clusterlist = [set(cl) for cl in clusters]
    clusterlist.sort(key=len)
    sanitised_clusters = []
    for _ in range(len(clusterlist)):
        cluster = clusterlist.pop(0)
        for cl in clusterlist:
            cluster.difference_update(cl)
        sanitised_clusters.append(cluster)
    return {frozenset(cluster) for cluster in sanitised_clusters}


def name_clusters(clusters: set) -> dict:
    """Give each cluster a unique name.

    Note: the first 26 names will be from A - Z. If more than 26 clusters are present,
    these will get names with two uppercase letters (AA - ZZ).

    Args:
        clusters: A set of different clusters. Each element is a list of clusters and
            their splits.

    Returns:
        A dictionary in the form {clustername0: clusters0, cluster_name1: cluster1}
    """
    cluster_names = list(string.ascii_uppercase)
    # Ensure a sufficiently long list of possible names, by extending (A-Z) with (AA-ZZ)
    cluster_names += [a + b for a, b in itertools.product(cluster_names, repeat=2)]
    clusters_list = list(clusters)

    # Ensure reproducable behavior, as sets are unordered.
    clusters_list.sort(
        key=lambda letters: sum(ord(c) for c in str(letters)), reverse=True
    )

    named_clusters = {}
    for cluster in clusters_list:
        name = cluster_names.pop(0)
        named_clusters[name] = set(cluster)

    return named_clusters


def create_renaming_dict(aligned_clusters: dict) -> dict[int, list[tuple[int, str]]]:
    """Create a dictionary that can be used to rename the clusters to the aligned names.

    Args:
        aligned_clusters: A dictionary containing the different splits, and the mapping
            of RGDR clusters to new names.

    Returns:
        A dictionary with the structure {split: [(old_name0, new_name0),
                                                 (old_name1, new_name1)]}.
    """
    inversed_names = {}
    for key in aligned_clusters:
        for el in aligned_clusters[key]:
            inversed_names[el] = key

    renaming_dict: dict[int, list[tuple[int, str]]] = {}
    for key in inversed_names:  # pylint: disable=consider-using-dict-items
        ky = key.split("_")
        split = int(ky[0])
        cluster = int(ky[1])
        if split in renaming_dict:
            apd = renaming_dict[split] + [(cluster, inversed_names[key])]
            renaming_dict[split] = apd
        else:
            renaming_dict[split] = [(cluster, inversed_names[key])]
    return renaming_dict


def ensure_unique_names(renaming_dict: dict[int, list[tuple[int, str]]]) -> dict:
    """Ensure that in every split, every cluster has a unique name.

    The function finds the non-unqiue names within each split, and will rename these by
    adding a number. For example, there are three clusters in the first split with the
    name "C". The new names will be "C", "C1" and "C2".

    If renaming_dict is the following:
        {0: [(-1, "A"), (1, "B")], 1: [(-1, "A"), (-2, "A")]}
    The renamed dictionary will be:
        {0: [(-1, "A1"), (1, "B")], 1: [(-1, "A1"), (-2, "A2")]}

    Args:
        renaming_dict: Renaming dictionary with non unique names.

    Returns:
        Renaming dictionary with only unique names
    """
    renamed_dict = deepcopy(renaming_dict)

    double_names_any: set[str] = set()
    for split in renamed_dict:
        cluster_old_names = [cl for cl, _ in renamed_dict[split]]
        cluster_new_names = [cl for _, cl in renamed_dict[split]]

        double_names = [
            x for x in set(cluster_new_names) if cluster_new_names.count(x) > 1
        ]

        for double_name in double_names:
            double_names_any.add(double_name)
            ndouble = cluster_new_names.count(double_name)
            for i in range(ndouble):
                el = cluster_new_names.index(double_name)
                cluster_new_names[el] = double_name + str(i + 1)
        renamed_dict[split] = list(zip(cluster_old_names, cluster_new_names))

    # If any new label was renamed in a different cluster, e.g. A -> A1, make sure that
    # any other "A" label is also renamed.
    for split, clusters in renamed_dict.items():
        for i, (old_name, new_name) in enumerate(clusters):
            if new_name in double_names_any:
                renamed_dict[split][i] = (old_name, f"{new_name}1")

    return renamed_dict


def _rename_datasets(
    rgdr_list: list["RGDR"], clustered_data: list[xr.DataArray], renaming_dict: dict
) -> list[xr.DataArray]:
    """Apply the renaming dictionary to the labels of the clustered data.

    Args:
        rgdr_list: List of RGDR objects that were used to fit and transform the data.
        clustered_data: List of the RGDR-transformed data. This can either be the
            training data or the test data.
        renaming_dict: Dictionary containing the mapping {old_label: new_label}

    Returns:
        A list of the input clustered data, with the labels renamed.
    """
    renamed = []
    for split, _ in enumerate(rgdr_list):
        # A copy is required to not modify labels of the input data
        data = copy(clustered_data[split])
        labels = data["cluster_labels"].values.astype("U2")
        for cl in renaming_dict[split]:
            if f"{cl[0]}" not in labels:
                break
            labels[labels == f"{cl[0]}"] = cl[1]
        data["cluster_labels"] = labels
        renamed.append(data)
    return renamed


def rename_labels(
    rgdr_list: list["RGDR"], clustered_data: list[xr.DataArray]
) -> list[xr.DataArray]:
    """Return a new object with renamed cluster labels aligned over different splits.

    To aid in users comparing the clustering over different splits, this function tries
    to match the clusters over different splits, and give clusters that are in the same
    region the same name (e.g. "A"). The clusters themselves are not changed, only the
    labels renamed.

    Args:
        rgdr_list: List of RGDR objects that were used to fit and transform the data.
        clustered_data: List of the RGDR-transformed datasets. This can either be the
            training data or the test data.

    Returns:
        A list of the input clustered data, with the labels renamed.
    """
    n_splits = len(rgdr_list)
    cluster_maps = [rgdr_list[split].cluster_map.reset_coords() for split in range(n_splits)]  # type: ignore

    cluster_map_ds = xr.concat(cluster_maps, dim="split")

    clusters = get_overlapping_clusters(cluster_map_ds["cluster_labels"])
    clusters = remove_subsets(clusters)
    clusters = remove_overlapping_clusters(clusters)
    cluster_names = name_clusters(clusters)

    renaming_dict = create_renaming_dict(cluster_names)
    renaming_dict = ensure_unique_names(renaming_dict)

    return _rename_datasets(rgdr_list, clustered_data, renaming_dict)



"""Tests for the s2spy.preprocess module."""
import numpy as np
import pytest
import scipy.signal
import xarray as xr
from s2spy import preprocess
from s2spy.preprocess import Preprocessor


TEST_FILE_PATH = "./tests/test_rgdr/test_data"

# pylint: disable=protected-access
# ruff: noqa: B018
# functions are called without variable assignment for testing the error message
# therefore need to suppress flake-burger B018


class TestPreprocessMethods:
    """Test preprocess methods."""

    # Define inputs as fixtures
    @pytest.fixture
    def raw_field(self):
        return xr.open_dataset(
            f"{TEST_FILE_PATH}/sst_daily_1979-2018_5deg_Pacific_175_240E_25_50N.nc"
        ).sel(
            time=slice("2010-01-01", "2011-12-31"),
            latitude=slice(40, 30),
            longitude=slice(180, 190),
        )

    def test_check_input_data_incorrect_type(self):
        dummy_data = np.ones((3, 3))
        with pytest.raises(ValueError):
            preprocess._check_input_data(dummy_data)

    def test_check_input_data_incorrect_dims(self, raw_field):
        dummy_dataarray = raw_field["sst"][0, ...]  # drop time dimension
        with pytest.raises(ValueError):
            preprocess._check_input_data(dummy_dataarray)

    def test_unknown_method_get_trend(self, raw_field):
        with pytest.raises(ValueError):
            preprocess._get_trend(raw_field, "unknown")

    def test_get_and_subtract_linear_trend(self, raw_field):
        expected = xr.apply_ufunc(
            scipy.signal.detrend,
            raw_field,
            input_core_dims=[["time"]],
            output_core_dims=[["time"]],
        ).transpose("time", ...)
        trend = preprocess._get_trend(raw_field, "linear")
        result = preprocess._subtract_trend(raw_field, "linear", trend)
        np.testing.assert_array_almost_equal(result["sst"], expected["sst"])

    def test_check_temporal_resolution(self):
        with pytest.raises(ValueError):
            preprocess._check_temporal_resolution("hourly")  # type: ignore

    def test_get_climatology_daily(self, raw_field):
        result = preprocess._get_climatology(raw_field, timescale="daily")
        expected = (
            raw_field["sst"].sel(time=slice("2010-01-01", "2010-12-31")).data
            + raw_field["sst"].sel(time=slice("2011-01-01", "2011-12-31")).data
        ) / 2
        np.testing.assert_array_almost_equal(result["sst"], expected)

    def test_get_climatology_weekly(self, raw_field):
        raw_field_weekly = raw_field.resample(time="W").mean()
        result = preprocess._get_climatology(raw_field_weekly, timescale="weekly")
        # need to consider the actual calendar week number for the expected climatology
        raw_field_weekly["time"] = raw_field_weekly["time"].dt.isocalendar().week
        expected = raw_field_weekly.groupby("time").mean()
        np.testing.assert_array_almost_equal(result["sst"], expected["sst"])

    def test_get_climatology_monthly(self, raw_field):
        raw_field_monthly = raw_field.resample(time="M").mean()
        result = preprocess._get_climatology(raw_field_monthly, timescale="monthly")
        expected = (
            raw_field_monthly["sst"].sel(time=slice("2010-01-01", "2010-12-31")).data
            + raw_field_monthly["sst"].sel(time=slice("2011-01-01", "2011-12-31")).data
        ) / 2
        np.testing.assert_array_almost_equal(result["sst"], expected)

    def test_get_climatology_wrong_timescale(self, raw_field):
        with pytest.raises(ValueError):
            preprocess._get_climatology(raw_field, timescale="hourly")  # type: ignore

    @pytest.mark.parametrize("timescale", ("weekly", "monthly"))
    def test_check_data_resolution_mismatch(self, raw_field, timescale):
        with pytest.warns(UserWarning):
            preprocess._check_data_resolution_match(raw_field, timescale)

    def test_check_data_resolution_match(self, raw_field):
        preprocess._check_data_resolution_match(raw_field, "daily")


class TestPreprocessor:
    """Test preprocessor."""

    @pytest.fixture
    def preprocessor(self):
        prep = preprocess.Preprocessor(
            rolling_window_size=25,
            timescale="daily",
            detrend="linear",
            subtract_climatology=True,
        )
        return prep

    @pytest.fixture(params=[None, 1])
    def preprocessor_no_rolling(self, request):
        prep = preprocess.Preprocessor(
            rolling_window_size=request.param,
            timescale="monthly",
            detrend=None,
            subtract_climatology=True,
        )
        return prep

    @pytest.fixture
    def preprocessor_no_climatology(self):
        prep = preprocess.Preprocessor(
            rolling_window_size=25,
            timescale="daily",
            detrend="linear",
            subtract_climatology=False,
        )
        return prep

    @pytest.fixture
    def preprocessor_no_detrend(self):
        prep = preprocess.Preprocessor(
            rolling_window_size=25,
            timescale="daily",
            detrend=None,
            subtract_climatology=True,
        )
        return prep

    @pytest.fixture
    def raw_field(self):
        return xr.open_dataset(
            f"{TEST_FILE_PATH}/sst_daily_1979-2018_5deg_Pacific_175_240E_25_50N.nc"
        ).sel(
            time=slice("2010-01-01", "2014-12-31"),  # including leap year
            latitude=slice(40, 30),
            longitude=slice(180, 190),
        )

    def test_init(self):
        prep = preprocess.Preprocessor(
            rolling_window_size=25,
            timescale="weekly",
            detrend="linear",
            subtract_climatology=True,
        )
        assert isinstance(prep, Preprocessor)

    def test_fit(self, preprocessor, raw_field):
        preprocessor.fit(raw_field)
        assert (
            preprocessor.climatology is not None
        )  # climatology with leap years are counted

    def test_fit_no_rolling(self, preprocessor_no_rolling, raw_field):
        preprocessor_no_rolling.fit(raw_field)
        assert preprocessor_no_rolling.climatology == preprocess._get_climatology(
            raw_field, timescale="daily"
        )

    def test_transform(self, preprocessor, raw_field):
        preprocessor.fit(raw_field)
        preprocessed_data = preprocessor.transform(raw_field)
        assert preprocessed_data is not None

    def test_transform_without_fit(self, preprocessor, raw_field):
        with pytest.raises(ValueError):
            preprocessor.transform(raw_field)

    def test_fit_and_transform_no_climatology(
        self, preprocessor_no_climatology, raw_field
    ):
        preprocessor_no_climatology.fit(raw_field)
        results = preprocessor_no_climatology.transform(raw_field)
        expected = preprocess._subtract_trend(
            raw_field, "linear", preprocessor_no_climatology.trend
        )
        assert results == expected

    def test_fit_and_transform_no_detrend(self, preprocessor_no_detrend, raw_field):
        preprocessor_no_detrend.fit(raw_field)
        results = preprocessor_no_detrend.transform(raw_field)
        expected = (
            raw_field.groupby("time.dayofyear") - preprocessor_no_detrend.climatology
        )
        assert results == expected

    def test_fit_and_transform_no_climatology_and_detrend(self, raw_field):
        prep = preprocess.Preprocessor(
            rolling_window_size=10,
            timescale="daily",
            detrend=None,
            subtract_climatology=False,
        )
        prep.fit(raw_field)
        results = prep.transform(raw_field)

        assert results == raw_field

    def test_fit_transform(self, preprocessor, raw_field):
        preprocessed_data = preprocessor.fit_transform(raw_field)
        assert preprocessed_data is not None

    def test_trend_property_not_fit(self, preprocessor):
        with pytest.raises(ValueError, match="The preprocessor has to be fit"):
            preprocessor.trend

    def test_trend_property_no_detrend(self, preprocessor_no_detrend, raw_field):
        preprocessor_no_detrend.fit(raw_field)
        with pytest.raises(ValueError, match="Detrending is set to `None`"):
            preprocessor_no_detrend.trend

    def test_climatology_property_not_fit(self, preprocessor):
        with pytest.raises(ValueError, match="The preprocessor has to be fit"):
            preprocessor.climatology

    def test_trend_property_no_climatology(self, preprocessor_no_climatology):
        with pytest.raises(ValueError, match="subtract_climatology is set to `False`"):
            preprocessor_no_climatology.climatology



"""Commonly used utility functions for s2spy."""
from typing import TypeVar
import numpy as np
import xarray as xr


XrType = TypeVar("XrType", xr.DataArray, xr.Dataset)


def weighted_groupby(
    ds: XrType, groupby: str, weight: str, method: str = "mean"
) -> XrType:
    """Apply a weighted reduction after a groupby call.

    xarray does not currently support combining `weighted` and `groupby`. An open PR
    adds supports for this functionality (https://github.com/pydata/xarray/pull/5480),
    but this branch was never merged.

    Args:
        ds: Data containing the coordinates or variables
            specified in the `groupby` and `weight` kwargs.
        groupby: Coordinate which should be used to make the groups.
        weight: Variable in the Dataset containing the weights that should be used.
        method: Method that should be used to reduce the dataset, by default
            'mean'. Supports any of xarray's builtin methods, e.g. 'median', 'min',
            'max'.

    Returns:
        Same as input: Dataset reduced using the `groupby` coordinate, using weights =
            based on `ds[weight]`.
    """
    groups = ds.groupby(groupby)

    # find stacked dim name
    group0 = list(groups)[0][1]
    dims = list(group0.dims)
    stacked_dims = [dim for dim in dims if "stacked_" in str(dim)]

    reduced_groups = [
        getattr(g.weighted(g[weight]), method)(dim=stacked_dims) for _, g in groups
    ]

    reduced_data: XrType = xr.concat(reduced_groups, dim=groupby)

    if isinstance(reduced_data, xr.DataArray):  # Add back the labels of the groupby dim
        reduced_data[groupby] = np.unique(ds[groupby])
    return reduced_data


def geographical_cluster_center(
    masked_data: xr.DataArray, reduced_data: xr.DataArray
) -> xr.DataArray:
    """Add the geographical centers to the clusters.

    Args:
        masked_data (xr.DataArray): Precursor data before being reduced to clusters,
            with the dimensions latitude and longitude, and cluster labels added.
        reduced_data (xr.DataArray): Data reduced to the clusters, to which the
            geographical centers will be added

    Returns:
        xr.DataArray: Reduced data with the latitude and longitude of the geographical
            centers added as coordinates of the cluster labels.
    """
    clusters = np.unique(masked_data["cluster_labels"])
    stacked_data = masked_data.stack(coords=("latitude", "longitude"))

    cluster_lats = np.zeros(clusters.shape)
    cluster_lons = np.zeros(clusters.shape)

    for i, cluster in enumerate(clusters):
        # Select only the grid cells within the cluster
        cluster_area = stacked_data["area"].where(
            stacked_data["cluster_labels"] == cluster
        )

        if "i_interval" in cluster_area.dims:
            cluster_area = cluster_area.dropna("i_interval", how="all")
        cluster_area = cluster_area.dropna("coords")

        # Area weighted mean to get the geographical center of the cluster
        cluster_lats[i] = cluster_area["latitude"].weighted(cluster_area).mean().item()
        cluster_lons[i] = cluster_area["longitude"].weighted(cluster_area).mean().item()

    reduced_data["latitude"] = ("cluster_labels", cluster_lats)
    reduced_data["longitude"] = ("cluster_labels", cluster_lons)

    return reduced_data


def intervals_subtract(intervals: list[int], n: int) -> list[int]:
    """Subtracts n from the interval indices, skipping 0."""
    if n < 0:
        raise ValueError("Lag values below 0 are not supported")

    lag_intervals = [i - n for i in intervals]
    # pylint: disable=chained-comparison
    return [
        i - 1 if (i <= 0 and j > 0) else i for i, j in zip(lag_intervals, intervals)
    ]



"""Preprocessor for s2spy workflow."""
import warnings
from typing import Literal
from typing import Union
import numpy as np
import scipy.stats
import xarray as xr


def _linregress(x: np.ndarray, y: np.ndarray) -> tuple[float, float]:
    """Calculate the slope and intercept between two arrays using scipy's linregress.

    Used to make linregress more ufunc-friendly.

    Args:
        x: First array.
        y: Second array.

    Returns:
        slope, intercept
    """
    slope, intercept, _, _, _ = scipy.stats.linregress(x, y)
    return slope, intercept


def _trend_linear(data: Union[xr.DataArray, xr.Dataset]) -> dict:
    """Calculate the linear trend over time.

    Args:
        data: The input data of which you want to know the trend.

    Returns:
        Dictionary containing the linear trend information (slope and intercept)
    """
    slope, intercept = xr.apply_ufunc(
        _linregress,
        data["time"].astype(float),
        data,
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[[], []],
        vectorize=True,
    )
    return {"slope": slope, "intercept": intercept}


def _subtract_linear_trend(data: Union[xr.DataArray, xr.Dataset], trend: dict):
    """Subtract a previously calclulated linear trend from (new) data."""
    return data - trend["intercept"] - trend["slope"] * (data["time"].astype(float))


def _get_trend(data: Union[xr.DataArray, xr.Dataset], method: str):
    """Calculate the trend, with a certain method. Only linear is implemented."""
    if method == "linear":
        return _trend_linear(data)
    raise ValueError(f"Unkown detrending method '{method}'")


def _subtract_trend(data: Union[xr.DataArray, xr.Dataset], method: str, trend: dict):
    """Subtract the previously calculated trend from (new) data. Only linear is implemented."""
    if method == "linear":
        return _subtract_linear_trend(data, trend)
    raise NotImplementedError


def _get_climatology(
    data: Union[xr.Dataset, xr.DataArray],
    timescale: Literal["monthly", "weekly", "daily"],
):
    """Calculate the climatology of timeseries data."""
    _check_data_resolution_match(data, timescale)
    if timescale == "monthly":
        climatology = data.groupby("time.month").mean("time")
    elif timescale == "weekly":
        climatology = data.groupby(data["time"].dt.isocalendar().week).mean("time")
    elif timescale == "daily":
        climatology = data.groupby("time.dayofyear").mean("time")
    else:
        raise ValueError("Given timescale is not supported.")

    return climatology


def _subtract_climatology(
    data: Union[xr.Dataset, xr.DataArray],
    timescale: Literal["monthly", "weekly", "daily"],
    climatology: Union[xr.Dataset, xr.DataArray],
):
    if timescale == "monthly":
        deseasonalized = data.groupby("time.month") - climatology
    elif timescale == "weekly":
        deseasonalized = data.groupby(data["time"].dt.isocalendar().week) - climatology
    elif timescale == "daily":
        deseasonalized = data.groupby("time.dayofyear") - climatology
    else:
        raise ValueError("Given timescale is not supported.")

    return deseasonalized


def _check_input_data(data: Union[xr.DataArray, xr.Dataset]):
    """Check the input data for compatiblity with the preprocessor.

    Args:
        data: Data to validate.

    Raises:
        ValueError: If the input data is of the wrong type.
        ValueError: If the input data does not have a 'time' dimension.
    """
    if not any(isinstance(data, dtype) for dtype in (xr.DataArray, xr.Dataset)):
        raise ValueError(
            "Input data has to be an xarray-DataArray or xarray-Dataset, "
            f"not {type(data)}"
        )
    if "time" not in data.dims:
        raise ValueError(
            "Analysis is done of the 'time' dimension, but the input data"
            f" only has dims: {data.dims}"
        )


def _check_temporal_resolution(
    timescale: Literal["monthly", "weekly", "daily"]
) -> Literal["monthly", "weekly", "daily"]:
    support_temporal_resolution = ["monthly", "weekly", "daily"]
    if timescale not in support_temporal_resolution:
        raise ValueError(
            "Given temporal resoltuion is not supported."
            "Please choose from 'monthly', 'weekly', 'daily'."
        )
    return timescale


def _check_data_resolution_match(
    data: Union[xr.DataArray, xr.Dataset],
    timescale: Literal["monthly", "weekly", "daily"],
):
    """Check if the temporal resolution of input is the same as given timescale."""
    timescale_dict = {
        "monthly": np.timedelta64(1, "M"),
        "weekly": np.timedelta64(1, "W"),
        "daily": np.timedelta64(1, "D"),
    }
    time_intervals = np.diff(data["time"].to_numpy())
    temporal_resolution = np.median(time_intervals).astype("timedelta64[D]")
    if timescale == "monthly":
        temporal_resolution = temporal_resolution.astype(int)
        min_days, max_days = (28, 31)
        if not max_days >= temporal_resolution >= min_days:
            warnings.warn(
                "The temporal resolution of data does not completely match "
                "the target timescale. Please check your input data.",
                stacklevel=1,
            )

    elif timescale in timescale_dict:
        if timescale_dict[timescale].astype("timedelta64[D]") != temporal_resolution:
            warnings.warn(
                "The temporal resolution of data does not completely match "
                "the target timescale. Please check your input data.",
                stacklevel=1,
            )


class Preprocessor:
    """Preprocessor for s2s data."""

    def __init__(  # noqa: PLR0913
        self,
        rolling_window_size: Union[int, None],
        timescale: Literal["monthly", "weekly", "daily"],
        rolling_min_periods: int = 1,
        subtract_climatology: bool = True,
        detrend: Union[str, None] = "linear",
    ):
        """Preprocessor for s2s data. Can detrend as well as deseasonalize.

        On calling `.fit(data)`, the preprocessor will:
         - Calculate the rolling mean of the input data.
         - Calculate and store the climatology of the rolling mean.
         - Calculate and store the trend of the rolling mean.

        When calling `.transform(data)`, the preprocessor will:
         - Remove the climatology from a copy of the data.
         - Remove the (stored) trend from this deseasonalized data.
         - Return the detrended and deseasonalized data.

        Args:
            rolling_window_size: The size of the rolling window that will be applied
                before calculating the trend and climatology. Setting this to None will
                skip this step.
            rolling_min_periods: The minimum number of periods within a rolling window.
                If higher than 1 (the default), NaN values will be present at the start
                and end of the preprocessed data.
            subtract_climatology (optional): If you want to calculate and remove the
                climatology of the data. Defaults to True.
            detrend (optional): Which method to use for detrending. Currently the only method
                supported is "linear". If you want to skip detrending, set this to None.
            timescale: Temporal resolution of input data.
        """
        self._window_size = rolling_window_size
        self._min_periods = rolling_min_periods
        self._detrend = detrend
        self._subtract_climatology = subtract_climatology
        if subtract_climatology:
            self._timescale = _check_temporal_resolution(timescale)

        self._climatology: Union[xr.DataArray, xr.Dataset]
        self._trend: dict
        self._is_fit = False

    def fit(self, data: Union[xr.DataArray, xr.Dataset]) -> None:
        """Fit this Preprocessor to input data.

        Args:
            data: Input data for fitting.
        """
        _check_input_data(data)
        if self._window_size not in [None, 1]:
            data_rolling = data.rolling(
                dim={"time": self._window_size},  # type: ignore
                min_periods=self._min_periods,
                center=True,
            ).mean()
        # TODO: give option to be a gaussian-like window, instead of a block.
        else:
            data_rolling = data

        if self._subtract_climatology:
            self._climatology = _get_climatology(data_rolling, self._timescale)

        if self._detrend is not None:
            if self._subtract_climatology:
                deseasonalized = _subtract_climatology(
                    data_rolling, self._timescale, self._climatology
                )
                self._trend = _get_trend(deseasonalized, self._detrend)
            else:
                self._trend = _get_trend(data_rolling, self._detrend)

        self._is_fit = True

    def transform(
        self, data: Union[xr.DataArray, xr.Dataset]
    ) -> Union[xr.DataArray, xr.Dataset]:
        """Apply the preprocessing steps to the input data.

        Args:
            data: Input data to perform preprocessing.

        Returns:
            Preprocessed data.
        """
        if not self._is_fit:
            raise ValueError(
                "The preprocessor has to be fit to data before a transform"
                " can be applied"
            )

        if self._subtract_climatology:
            d = _subtract_climatology(data, self._timescale, self._climatology)
        else:
            d = data

        if self._detrend is not None:
            return _subtract_trend(d, self._detrend, self.trend)

        return d

    def fit_transform(
        self, data: Union[xr.DataArray, xr.Dataset]
    ) -> Union[xr.DataArray, xr.Dataset]:
        """Fit this Preprocessor to input data, and then apply the steps to the data.

        Args:
            data: Input data for fit and transform.

        Returns:
            Preprocessed data.
        """
        self.fit(data)
        return self.transform(data)

    @property
    def trend(self) -> dict:
        """Return the stored trend (dictionary)."""
        if not self._detrend:
            raise ValueError("Detrending is set to `None`, so no trend is available")
        if not self._is_fit:
            raise ValueError(
                "The preprocessor has to be fit to data before the trend"
                " can be requested."
            )
        return self._trend

    @property
    def climatology(self) -> Union[xr.DataArray, xr.Dataset]:
        """Return the stored climatology data."""
        if not self._subtract_climatology:
            raise ValueError(
                "`subtract_climatology is set to `False`, so no climatology "
                "data is available"
            )
        if not self._is_fit:
            raise ValueError(
                "The preprocessor has to be fit to data before the"
                " climatology can be requested."
            )
        return self._climatology



import lilio
import numpy as np
import pytest
import xarray as xr
from sklearn.model_selection import ShuffleSplit
import s2spy
import s2spy.rgdr
from s2spy import RGDR
from s2spy.rgdr import label_alignment


TEST_FILE_PATH = "./tests/test_rgdr/test_data"


class TestAlignmentSubfunctions:
    labels = np.array(
        [
            [
                [0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, -2, -2, -2, 0, 0],
                [-3, -3, -3, -3, 0, 0, 0, -2, -2, -2, 0, 0, 0],
            ],
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0],
            ],
            [
                [0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2, -2, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, -2, -2, -2, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, -2, -2, -2, 0, 0],
                [-3, 0, 0, 0, 0, 0, -4, 0, -2, -2, 0, -2, 0],
            ],
            [
                [0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 0],
                [0, 0, 0, 0, 0, -2, -2, -2, -2, -2, -2, 0, 0],
            ],
        ]
    )

    cluster_labels = xr.DataArray(labels, dims=("split", "latitude", "longitude"))

    expected_set_of_sets = {
        frozenset({"0_-3", "2_-3"}),
        frozenset({"0_-1", "2_-1", "3_-1"}),
        frozenset({"0_-2", "1_-1", "2_-2", "3_-2"}),
        frozenset({"0_-2", "1_-1", "2_-2", "2_-4", "3_-2"}),
        frozenset({"1_1", "2_1", "3_1"}),
        frozenset({"1_-1", "2_-4", "3_-2"}),
    }

    expected_reduced_sets = {
        frozenset({"0_-3", "2_-3"}),
        frozenset({"0_-1", "2_-1", "3_-1"}),
        frozenset({"0_-2", "1_-1", "2_-2", "2_-4", "3_-2"}),
        frozenset({"1_1", "2_1", "3_1"}),
    }

    def test_df_to_full_sets(self):
        set_of_sets = label_alignment.get_overlapping_clusters(
            self.cluster_labels, min_overlap=0.10
        )
        assert self.expected_set_of_sets == set_of_sets

    def test_remove_subsets(self):
        reduced_set_of_sets = label_alignment.remove_subsets(self.expected_set_of_sets)
        assert self.expected_reduced_sets == reduced_set_of_sets

    def test_remove_overlapping_clusters(self):
        dummy_input = {frozenset({"0_-1", "1_-2"}), frozenset({"0_-1", "1_-1", "2_-1"})}
        expected = {frozenset({"1_-2"}), frozenset({"0_-1", "1_-1", "2_-1"})}
        assert expected == label_alignment.remove_overlapping_clusters(dummy_input)

    def test_name_clusters(self):
        dummy_input = {frozenset({"0_-1", "1_-2"}), frozenset({"0_-1", "1_-1", "2_-1"})}
        expected = {
            "A": {"0_-1", "1_-1", "2_-1"},
            "B": {"0_-1", "1_-2"},
        }
        assert expected == label_alignment.name_clusters(dummy_input)

    def test_create_renaming_dict(self):
        dummy_cluster_names = {
            "A": {"0_-1", "1_-1", "2_-1"},
            "B": {"0_-2", "1_-2"},
        }
        expected_dict = {
            0: [(-1, "A"), (-2, "B")],
            1: [(-1, "A"), (-2, "B")],
            2: [(-1, "A")],
        }
        assert expected_dict == label_alignment.create_renaming_dict(
            dummy_cluster_names
        )

    def test_ensure_unique_names(self):
        dummy_names = {0: [(-1, "B")], 1: [(-2, "B"), (-1, "B")], 2: [(-1, "B")]}
        expected = {0: [(-1, "B1")], 1: [(-2, "B1"), (-1, "B2")], 2: [(-1, "B1")]}
        assert expected == label_alignment.ensure_unique_names(dummy_names)


@pytest.fixture(scope="class")
def dummy_calendar():
    return lilio.daily_calendar(anchor="08-01", length="30d")


@pytest.fixture(scope="module")
def raw_target():
    return xr.open_dataset(f"{TEST_FILE_PATH}/tf5_nc5_dendo_80d77.nc").sel(cluster=3)


@pytest.fixture(scope="module")
def raw_field():
    return xr.open_dataset(
        f"{TEST_FILE_PATH}/sst_daily_1979-2018_5deg_Pacific_175_240E_25_50N.nc"
    )


@pytest.fixture(scope="class")
def example_field(raw_field, dummy_calendar):
    cal = dummy_calendar.map_to_data(raw_field)
    return lilio.resample(cal, raw_field).sst


@pytest.fixture(scope="class")
def example_target(raw_target, raw_field, dummy_calendar):
    cal = dummy_calendar.map_to_data(raw_field)
    return lilio.resample(cal, raw_target).ts


def test_alignment_example(example_field, example_target):
    seed = 1  # same 'randomness'
    n_splits = 4

    shufflesplit = ShuffleSplit(n_splits=n_splits, test_size=0.25, random_state=seed)
    cv = lilio.traintest.TrainTestSplit(shufflesplit)

    rgdrs = [
        RGDR(target_intervals=[1], lag=5, eps_km=800, alpha=0.10, min_area_km2=0)
        for _ in range(n_splits)
    ]

    target_timeseries_splits = []
    precursor_field_splits = []
    for x_train, _, _, _ in cv.split(example_field, y=example_target):
        target_timeseries_splits.append(
            example_target.sel(anchor_year=x_train["anchor_year"].values)
        )
        precursor_field_splits.append(
            example_field.sel(anchor_year=x_train["anchor_year"].values)
        )

    clustered_precursors = [
        rgdrs[i].fit_transform(precursor_field_splits[i], target_timeseries_splits[i])
        for i in range(n_splits)
    ]

    aligned_precursors = s2spy.rgdr.label_alignment.rename_labels(
        rgdrs, clustered_precursors
    )

    expected = [["A1"], ["A1", "A2"], ["A1"], ["A1"]]
    clusters_list = [sorted(el.cluster_labels.values) for el in aligned_precursors]

    assert expected == clusters_list
